#
# Copyright IBM Corp. 2024 - 2024
# SPDX-License-Identifier: MIT
#

from quackling.llama_index.readers.docling_json_reader import DoclingJSONReader  # noqa
from quackling.llama_index.readers.docling_pdf_reader import DoclingPDFReader  # noqa
